/*
 * Parse.l - basic lexical analyzer
 *
 *	Vax/Basic to C++ lexical analyzer
 *
 * Description:
 *
 *	This procedure is used to break Vax/Basic source code
 *	into individual tokens, when are passed on to 'parse.y'.
 *
 * History:
 *
 *	01.01.00 - 08/31/1996 - Kevin Handy
 *		Set version to 1.1, and install on internet.
 *
 *	01.01.01 - 09/04/1996 - Kevin Handy
 *		Added "BLOCK" keyword.
 *
 *	01.01.02 - 10/09/1996 - Kevin Handy
 *		Added BAS_V_PREDEF information.
 *
 *	01.01.03 - 10/23/1996 - Kevin Handy
 *		Added HFLOAT symbol.
 *
 *	01.03.01 - 03/16/1997 - Kevin Handy
 *		Moved 'mangle_string' from 'nodes.cc' to here.
 *		Fixed bug in 'xxx' quoted strings.
 *
 *	01.03.02 - 03/20/1997 - Kevin Handy
 *		Handle '0'C as a special case (confuses C strings
 *		if translated as "\0")
 *
 *	01.03.03 - 04/02/1997 - Kevin Handy
 *		Added "CLUSTERSIZE" keyword.
 *		Handle '[]' like '()'. RSTS Basic+ seems to work so.
 *
 *	01.03.04 - 04/21/1997 - Kevin Handy
 *		Added several more BASIC constants.
 *
 *	01.03.05 - 06/09/1997 - Kevin Handy
 *		Added keyword 'explicit' and '%library'
 *
 *	01.04.00 - 06/16/1997 - Kevin Handy
 *		Set version to 1.4
 *
 *	01.04.01 - 07/09/1997 - Kevin Handy
 *		Added 'MAT'.
 */
%x incl
%x xremark

%{
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "basic.h"
#include "nodes.h"
#include "parse.h"

int xline = 1;
int include_stack_pointer = 0;
Node* CommentList;

static const int MAX_INCLUDE_DEPTH = 10;
YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
static int include_line[MAX_INCLUDE_DEPTH];
char* include_name[MAX_INCLUDE_DEPTH] = {NULL};
static void StartInclude(char* filename);
static char *mangle_string(char *Text);
static void mangle_string_value(char *ThisPart, int Base, char Type);

#ifndef max
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) > (b)) ? (a) : (b))
#endif

static inline int SetReturn(int x)
{
	yylval = new Node(x, yytext, include_stack_pointer);
	return(x);
}

static inline int SetStringReturn(int x)
{
	char* MangleValue = mangle_string(yytext);
	yylval = new Node(x, MangleValue, include_stack_pointer);
	delete MangleValue;
	return(x);
}

static inline int SetNullReturn(int x)
{
	yylval = new Node(x, NULL, include_stack_pointer);
	return(x);
}

static inline void SetComment(int x)
{
	yylval = new Node(x, yytext, include_stack_pointer);

	if (CommentList == NULL)
	{
		CommentList = yylval;
	}
	else
	{
		CommentList->DownLink(yylval);
	}
}

static inline void CountLine()
{
	xline++;
}

%}


%%
	/*
	 * White-space characters
	 */
[\n\f]			{ CountLine(); return('\n'); }
"&"[\n\f]		CountLine();
[ \t]+			;
"\""([^\"]*)"\""	return SetStringReturn(BAS_V_TEXTSTRING);
"'"([^']*)"'"		return SetStringReturn(BAS_V_TEXTSTRING);
[DBOX]?"\""([^\"]*)"\""[BWLFDGHP]	return SetStringReturn(BAS_V_INTEGER);
[DBOX]?"'"([^']*)"'"[BWLFDGHP]	return SetStringReturn(BAS_V_INTEGER);
[DBOX]?"\""([^\"]*)"\"C"	return SetStringReturn(BAS_V_TEXTSTRING);
[DBOX]?"'"([^']*)"'C"		return SetStringReturn(BAS_V_TEXTSTRING);

"!"[^\n\f\!]*"!"	{ SetComment(BAS_S_REMARK); }
"!"[^\n\f\!]*/"&"[\n\f]	{ SetComment(BAS_S_REMARK); }
"!"[^\n\f\!]*/[\n\f]	{ SetComment(BAS_S_REMARK); }
"REM"			{ BEGIN(xremark); }

<xremark>[^\n\f]*/"&"[\n\f]	{ SetComment(BAS_S_REMARK); BEGIN(INITIAL); }
<xremark>[^\n\f]*/[\n\f]	{ SetComment(BAS_S_REMARK); BEGIN(INITIAL); }

	/*
	 * Single character operations
	 */
"(" |
")" |
"*" |
"/" |
"\\" |
"+" |
"-" |
":" |
";" |
"," |
"#" |
"^" |
"=" |
"==" |
"<" |
">"		return SetReturn(yytext[0]);

"["		return SetReturn('(');
"]"		return SetReturn(')');

"**"		return SetNullReturn('^');
"::"		return SetNullReturn(BAS_X_STRREF);
"><" |
"<>"		return SetNullReturn(BAS_X_NEQ);
">=" |
"=>"		return SetNullReturn(BAS_X_GE);
"<=" |
"=<"		return SetNullReturn(BAS_X_LE);

	/*
	 * Statements (usually)
	 */
"ACCESS"	return SetNullReturn(BAS_S_ACCESS);
"ALLOW"		return SetNullReturn(BAS_S_ALLOW);
"ALTERNATE"	return SetNullReturn(BAS_S_ALTERNATE);
"AND"		return SetNullReturn(BAS_S_AND);
"ANY"		return SetNullReturn(BAS_S_ANY);
"APPEND"	return SetNullReturn(BAS_S_APPEND);
"BACK"		return SetNullReturn(BAS_S_BACK);
"BLOCK"		return SetNullReturn(BAS_S_BLOCK);
"BLOCKSIZE"	return SetNullReturn(BAS_S_BLOCKSIZE);
"BUCKETSIZE"	return SetNullReturn(BAS_S_BUCKETSIZE);
"BUFFER"	return SetNullReturn(BAS_S_BUFFER);
"BY"		return SetNullReturn(BAS_S_BY);
"CALL"		return SetNullReturn(BAS_S_CALL);
"CASE"[ \t]*"ELSE"	return SetNullReturn(BAS_N_CASEELSE);
"CASE"		return SetNullReturn(BAS_S_CASE);
"CHAIN"		return SetNullReturn(BAS_S_CHAIN);
"CHANGES"	return SetNullReturn(BAS_S_CHANGES);
"CHANGE"	return SetNullReturn(BAS_S_CHANGE);
"CLOSE"		return SetNullReturn(BAS_S_CLOSE);
"CLUSTERSIZE"	return SetNullReturn(BAS_S_CLUSTERSIZE);
"COM" |
"COMMON"	return SetNullReturn(BAS_S_COMMON);
"CONSTANT"	return SetNullReturn(BAS_S_CONSTANT);
"CONTIGUOUS"	return SetNullReturn(BAS_S_CONTIGUOUS);
"COUNT"		return SetNullReturn(BAS_S_COUNT);
"DATA"		return SetNullReturn(BAS_S_DATA);
"DECIMAL"	return SetNullReturn(BAS_S_DECIMAL);
"DECLARE"	return SetNullReturn(BAS_S_DECLARE);
"DEF"		return SetNullReturn(BAS_S_DEF);
"DEFAULTNAME"	return SetNullReturn(BAS_S_DEFAULTNAME);
"DELETE"	return SetNullReturn(BAS_S_DELETE);
"DESC"		return SetNullReturn(BAS_S_DESC);
"DIM" |
"DIMENSION"	return SetNullReturn(BAS_S_DIM);
"DUPLICATES"	return SetNullReturn(BAS_S_DUPLICATES);
"END"		return SetNullReturn(BAS_S_END);
"END"[ \t]*"IF"	return SetNullReturn(BAS_N_ENDIF);
"EQ"		return SetNullReturn(BAS_S_EQ);
"EQV"		return SetNullReturn(BAS_S_EQV);
"ERL"		{ NeedErrorH = 1; return SetReturn(BAS_S_ERL); }
"ERR"		{ NeedErrorH = 1; return SetReturn(BAS_S_ERR); }
"ERROR"		return SetNullReturn(BAS_S_ERROR);
"EXIT"		return SetNullReturn(BAS_S_EXIT);
"EXPLICIT"	return SetNullReturn(BAS_S_EXPLICIT);
"EXTEND"	return SetNullReturn(BAS_S_EXTEND);
"EXTENDSIZE"	return SetNullReturn(BAS_S_EXTENDSIZE);
"EXTERNAL"	return SetNullReturn(BAS_S_EXTERNAL);
"FIELD"		return SetReturn(BAS_S_FIELD);
"FILESIZE"	return SetNullReturn(BAS_S_FILESIZE);
"FILL"		return SetReturn(BAS_S_FILL);
"FILL%"		return SetReturn(BAS_S_FILL);
"FILL$"		return SetReturn(BAS_S_FILL);
"FIND"		return SetNullReturn(BAS_S_FIND);
"FIXED"		return SetReturn(BAS_S_FIXED);
"FNEND"		return SetNullReturn(BAS_S_FNEND);
"FNEXIT"	return SetNullReturn(BAS_S_FNEXIT);
"FOR"		return SetNullReturn(BAS_S_FOR);
"FORTRAN"	return SetReturn(BAS_S_FORTRAN);
"FREE"		return SetNullReturn(BAS_S_FREE);
"FROM"		return SetNullReturn(BAS_S_FROM);
"FUNCTION"	return SetNullReturn(BAS_S_FUNCTION);
"FUNCTIONEND"	return SetNullReturn(BAS_S_FUNCTIONEND);
"FUNCTIONEXIT"	return SetNullReturn(BAS_S_FUNCTIONEXIT);
"GE"		return SetNullReturn(BAS_S_GE);
"GET"		return SetNullReturn(BAS_S_GET);
"GO"		return SetNullReturn(BAS_S_GO);
"GOSUB"		return SetNullReturn(BAS_S_GOSUB);
"GOTO"		return SetNullReturn(BAS_S_GOTO);
"GT"		return SetNullReturn(BAS_S_GT);
"IF"		return SetNullReturn(BAS_S_IF);
"IMP"		return SetNullReturn(BAS_S_IMP);
"INDEXED"	return SetReturn(BAS_S_INDEXED);
"INPUT"		return SetNullReturn(BAS_S_INPUT);
"ITERATE"	return SetNullReturn(BAS_S_ITERATE);
"KEY"		return SetNullReturn(BAS_S_KEY);
"KILL"		return SetNullReturn(BAS_S_KILL);
"LET"		return SetNullReturn(BAS_S_LET);
"LINPUT"	return SetNullReturn(BAS_S_LINPUT);
"LIST"		return SetReturn(BAS_S_LIST);
"LSET"		return SetNullReturn(BAS_S_LSET);
"MAP"		return SetNullReturn(BAS_S_MAP);
"MARGIN"	return SetNullReturn(BAS_S_MARGIN);
"MAT"		return SetNullReturn(BAS_S_MAT);
"MODE"		return SetNullReturn(BAS_S_MODE);
"MODIFY"	return SetReturn(BAS_S_MODIFY);
"MOVE"		return SetNullReturn(BAS_S_MOVE);
"NAME"		return SetNullReturn(BAS_S_NAME);
"NEXT"		return SetNullReturn(BAS_S_NEXT);
"NOCHANGES"	return SetNullReturn(BAS_S_NOCHANGES);
"NODUPLICATES"	return SetNullReturn(BAS_S_NODUPLICATES);
"NONE"		return SetReturn(BAS_S_NONE);
"NOSPAN"	return SetReturn(BAS_S_NOSPAN);
"NOT"		return SetNullReturn(BAS_S_NOT);
"NUM"		return SetNullReturn(BAS_S_NUM);
"NUM2"		return SetNullReturn(BAS_S_NUM2);
"ON"		return SetReturn(BAS_S_ON);
"OPEN"		return SetNullReturn(BAS_S_OPEN);
"OPTION"	return SetNullReturn(BAS_S_OPTION);
"OR"		return SetNullReturn(BAS_S_OR);
"OTHERWISE"	return SetNullReturn(BAS_S_OTHERWISE);
"OUTPUT"	return SetNullReturn(BAS_S_OUTPUT);
"ORGANIZATION"	return SetNullReturn(BAS_S_ORGANIZATION);
"PRIMARY"	return SetNullReturn(BAS_S_PRIMARY);
"PRINT"		return SetNullReturn(BAS_S_PRINT);
"PUT"		return SetNullReturn(BAS_S_PUT);
"RANDOM"	{ NeedTimeH = 1; NeedBasicFun = 1; return SetNullReturn(BAS_S_RANDOM); }
"RANDOMIZE"	{ NeedTimeH = 1; NeedBasicFun = 1; return SetNullReturn(BAS_S_RANDOM); }
"CTRLC"		return SetNullReturn(BAS_S_CTRLC);
"DET"		return SetNullReturn(BAS_S_DET);
"RCTLC"		return SetNullReturn(BAS_S_RCTLC);
"RCTLO"		return SetNullReturn(BAS_S_RCTLO);
"READ"		return SetNullReturn(BAS_S_READ);
"RECORD"	return SetNullReturn(BAS_S_RECORD);
"RECORDTYPE"	return SetNullReturn(BAS_S_RECORDTYPE);
"RECORDSIZE"	return SetNullReturn(BAS_S_RECORDSIZE);
"RECOUNT"	return SetNullReturn(BAS_S_RECOUNT);
"REF"		return SetNullReturn(BAS_S_REF);
"REGARDLESS"	return SetNullReturn(BAS_S_REGARDLESS);
"RELATIVE"	return SetReturn(BAS_S_RELATIVE);
"RESET"		return SetNullReturn(BAS_S_RESET);
"RESTORE"	return SetNullReturn(BAS_S_RESTORE);
"RESUME"	return SetNullReturn(BAS_S_RESUME);
"RETURN"	return SetNullReturn(BAS_S_RETURN);
"RSET"		return SetNullReturn(BAS_S_RSET);
"SCRATCH"	return SetReturn(BAS_S_SCRATCH);
"SELECT"	return SetNullReturn(BAS_S_SELECT);
"SEQUENTIAL"	return SetReturn(BAS_S_SEQUENTIAL);
"SIZE"		return SetNullReturn(BAS_S_SIZE);
"SLEEP"		return SetNullReturn(BAS_S_SLEEP);
"SPAN"		return SetReturn(BAS_S_SPAN);
"STATUS"	return SetNullReturn(BAS_S_STATUS);
"STEP"		return SetNullReturn(BAS_S_STEP);
"STOP"		return SetNullReturn(BAS_S_STOP);
"STREAM"	return SetReturn(BAS_S_STREAM);
"SUB"		return SetNullReturn(BAS_S_SUB);
"SUBEND"	return SetNullReturn(BAS_S_SUBEND);
"SUBEXIT"	return SetNullReturn(BAS_S_SUBEXIT);
"TEMPORARY"	return SetReturn(BAS_S_TEMPORARY);
"TYPE"		return SetNullReturn(BAS_S_TYPE);
"UNDEFINED"	return SetReturn(BAS_S_UNDEFINED);
"UNLESS"	return SetNullReturn(BAS_S_UNLESS);
"UNLOCK"	return SetNullReturn(BAS_S_UNLOCK);
"UNTIL"		return SetNullReturn(BAS_S_UNTIL);
"UPDATE"	return SetNullReturn(BAS_S_UPDATE);
"VALUE"		return SetNullReturn(BAS_S_VALUE);
"VARIABLE"	return SetReturn(BAS_S_VARIABLE);
"VARIANT"	return SetNullReturn(BAS_S_VARIANT);
"VIRTUAL"	return SetNullReturn(BAS_S_VIRTUAL);
"WAIT"		return SetNullReturn(BAS_S_WAIT);
"WHILE"		return SetNullReturn(BAS_S_WHILE);
"WINDOWSIZE"	return SetNullReturn(BAS_S_WINDOWSIZE);
"WRITE"		return SetNullReturn(BAS_S_WRITE);
"XOR"		return SetNullReturn(BAS_S_XOR);


"AS"		return SetNullReturn(BAS_S_AS);
"BYTE"		return SetReturn(BAS_S_BYTE);
"DOUBLE"	return SetReturn(BAS_S_DOUBLE);
"ELSE"		return SetNullReturn(BAS_S_ELSE);
"FILE"		return SetNullReturn(BAS_S_FILE);
"GFLOAT"	return SetReturn(BAS_S_GFLOAT);
"HFLOAT"	return SetReturn(BAS_S_HFLOAT);
"INTEGER"	return SetReturn(BAS_S_INTEGER);
"LINE"		return SetReturn(BAS_S_LINE);
"LONG"		return SetReturn(BAS_S_LONG);
"REAL"		return SetReturn(BAS_S_REAL);
"RFA"		{ NeedRFA = 1; return SetReturn(BAS_S_RFA); }
"SINGLE"	return SetReturn(BAS_S_SINGLE);
"STRING"	return SetReturn(BAS_S_STRING);
"THEN"		return SetNullReturn(BAS_S_THEN);
"TO"		return SetNullReturn(BAS_S_TO);
"USING"		return SetNullReturn(BAS_S_USING);
"WORD"		return SetReturn(BAS_S_WORD);


	/*
	 * Functions (usually)
	 */
"RND"		{ NeedStdlibH = 1; NeedBasicFun = 1; return SetReturn(BAS_V_NAME); }
"INSTR"		{ NeedStringH = 1; NeedBasicFun = 1; return SetReturn(BAS_V_FUNCTION); }

"ABS" |
"ABS%" |
"ATN" |
"COS" |
"EXP" |
"FIX" |
"INT" |
"LOG10" |
"LOG" |
"SIN" |
"SQRT" |
"SQR" |
"TAN" |
"VAL%" |
"VAL"		{ NeedMathH = 1; return SetReturn(BAS_V_FUNCTION); }

"ASC" |
"ASCII" |
"BUFSIZ" |
"CCPOS" |
"CHR$" |
"COMP%" |
"CVT$$" |
"CVT$%" |
"CVT%$" |
"CVT$F" |
"CVTF$" |
"DATE$" |
"DIF$" |
"ECHO" |
"EDIT$" |
"ERT$" |
"FORMAT$" |
"GETRFA" |
"MAR" |
"NUM$" |
"NUM1$" |
"ONECHR" |
"PLACE$" |
"RAD$" |
"RCTRLO" |
"SGN" |
"SPACE$" |
"STR$" |
"STRING$" |
"SWAP%" |
"SYS" |
"TAB" |
"TIME$" |
"TIME" |
"TRM$" |
"XLATE"		return SetReturn(BAS_V_FUNCTION);

"LEFT$" |
"LEFT" |
"LEN" |
"LOC" |
"MID$" |
"MID" |
"POS" |
"RIGHT$" |
"RIGHT" |
"SEG$"		{ NeedStringH = 1; return SetReturn(BAS_V_FUNCTION); }

"BEL" |
"BS" |
"HT" |
"LF" |
"VT" |
"FF" |
"CR" |
"SO" |
"SI" |
"ESC" |
"SP" |
"DEL" |
"PI"		return SetReturn(BAS_V_PREDEF);


	/*
	 * Percent Sign Commands
	 */
"%INCLUDE"	{ BEGIN(incl); return SetReturn(BAS_P_INCLUDE); }
"%FROM"		return SetReturn(BAS_P_FROM);
"%CDD"		return SetReturn(BAS_P_CDD);
"%PAGE"		;
"%TITLE"	return SetReturn(BAS_P_TITLE);
"%SBTTL"	return SetReturn(BAS_P_SBTTL);
"%IDENT"	return SetReturn(BAS_P_IDENT);
"%LIBRARY"	return SetReturn(BAS_P_LIBRARY);

"%CROSS" |
"%NOCROSS" |
"%LIST" |
"%NOLIST"	SetComment(BAS_S_REMARK);

	/*
	 * These lines are used to trap an include file name
	 */
<incl>[ \t]+		;
<incl>[\n\f]		{ BEGIN(INITIAL); CountLine(); return('\n'); }
<incl>"%FROM"		{ BEGIN(INITIAL); return SetReturn(BAS_P_FROM); }
<incl>"\""([^\"]*)"\""	{ char* MangleValue = mangle_string(yytext);
				yylval = new Node(BAS_V_TEXTSTRING,
				MangleValue, include_stack_pointer);
				StartInclude(yytext); BEGIN(INITIAL);
				delete MangleValue;
				return SetReturn(BAS_V_TEXTSTRING); }
<incl>"'"([^']*)"'"	{ char* MangleValue = mangle_string(yytext);
				yylval = new Node(BAS_V_TEXTSTRING,
				MangleValue, include_stack_pointer);
				StartInclude(yytext); BEGIN(INITIAL);
				delete MangleValue;
				return SetReturn(BAS_V_TEXTSTRING); }

	/*
	 * Numbers, variables, etc.
	 */
(([0-9]+"."[0-9]*)|("."[0-9]+))([ED]"-"?[0-9]+)?	return SetReturn(BAS_V_FLOAT);
[0-9]+[ED]"-"?[0-9]+	return SetReturn(BAS_V_FLOAT);
[0-9]+		return SetReturn(BAS_V_INTEGER);
[0-9]+[%]	return SetReturn(BAS_V_INT);
[A-Z][A-Z0-9._$]*[%]?		return SetReturn(BAS_V_NAME);
[A-Z][A-Z0-9_]*":"/[^:]		return SetReturn(BAS_V_LABEL);

	/*
	 * End of file (try to back up one include level)
	 */
<<EOF>>		{ if (--include_stack_pointer < 0)
			{ include_stack_pointer = 0; yyterminate(); }
			else
			{ yy_switch_to_buffer(
			include_stack[include_stack_pointer]);
			xline = include_line[include_stack_pointer];
			if (PositionDump){cerr << "Leaving Include " <<
			include_name[include_stack_pointer] << endl; }
			delete include_name[include_stack_pointer];
			}
		}
%%

static void StartInclude(char* FileName)
{
	char UseName[64];	// File name to actually use
	FILE* NewChannel;	// File Open Info

	//
	// Strip off quotes (We KNOW that we have them)
	//
	assert(FileName[0] != 0);
	assert(FileName[1] != 0);
	strcpy(UseName, FileName + 1);
	UseName[strlen(UseName) - 1] = '\0';

	//
	// Can we handle the include?
	//
	if (include_stack_pointer >= MAX_INCLUDE_DEPTH)
	{
		cerr << "Includes Nested Too Deep (" << UseName << ")" << endl;
		return;
	}

	//
	// Try to open up the file
	//
	NewChannel = fopen(UseName, "r");
	if (NewChannel == 0)
	{
		//
		// Try to lose any logicals on front of file name
		//
		char* cindex = strchr(UseName, ':');
		if (cindex != 0)
		{
			NewChannel = fopen(cindex + 1, "r");
			if (NewChannel == 0)
			{
				cerr << "Unable to open (" << UseName << ")" << endl;
				return;
			}
		}
		else
		{
			cerr << "Unable to open (" << UseName << ")" << endl;
			return;
		}
	}

	//
	// Switch over to new buffer
	//
	include_stack[include_stack_pointer] = YY_CURRENT_BUFFER;
	include_line[include_stack_pointer] = xline;
	include_name[include_stack_pointer] = new char[strlen(UseName)+1];
	strcpy(include_name[include_stack_pointer], UseName);
	xline = 1;
	include_stack_pointer++;

	yy_switch_to_buffer(yy_create_buffer(NewChannel, YY_BUF_SIZE));
	if (PositionDump) { cerr << "Including (" << UseName << ")" << endl;};
}

/*******************************************************************************

Local Function:

	mangle_string()

Title:

	Translates a quoted string from Basic format to C++ format.

Description:

	This function processes a quoted string, changing the
	format from the way Basic does it to the way C++ does
	it.

Bugs:


*******************************************************************************/

static char *mangle_string(char *Text)
{
	char *ThisPart;
	char *ThisPtr;
	char qmark;
	int Base = 10;

	assert(Text != NULL);

	//
	// Create area to work in
	// (Probibly more than we will need)
	//
	ThisPart = new char[max(14, strlen(Text) * 2)];
	ThisPtr = ThisPart;

	//
	// Handle any possible bases on front of string
	//
	switch (*Text)
	{
	case 'd':
	case 'D':
		Base = 10;
		Text++;
		break;
	case 'b':
	case 'B':
		Base = 2;
		Text++;
		break;
	case 'o':
	case 'O':
		Base = 8;
		Text++;
		break;
	case 'x':
	case 'X':
		Base = 16;
		Text++;
		break;
	}

	//
	// Determine quote character used (Better start with a quote
	// or were in deep ****)
	//
	qmark = '\0';

	//
	// Process all the characters in original string
	//
	while (*Text)
	{
		switch (*Text)
		{
		//
		// Slash must be doubled
		//
		case '\\':
			*ThisPtr++ = '\\';
			*ThisPtr++ = '\\';
			break;

		//
		// Change (audible bell) character to (\a)
		//
		case '\a':
			*ThisPtr++ = '\\';
			*ThisPtr++ = 'a';
			break;

		//
		// Change (backspace) character to (\b)
		//
		case '\b':
			*ThisPtr++ = '\\';
			*ThisPtr++ = 'b';
			break;

		//
		// Change (tab) character to (\t)
		//
		case '\t':
			*ThisPtr++ = '\\';
			*ThisPtr++ = 't';
			break;

		//
		// Change (vertical tab) character to (\v)
		//
		case '\v':
			*ThisPtr++ = '\\';
			*ThisPtr++ = 'v';
			break;

		//
		// Single quote mark
		//
		case '\'':
			switch(qmark)
			{
			case '\'':
				qmark = '\0';
				*ThisPtr++ = '"';
				break;
			case '\0':
				qmark = '\'';
				*ThisPtr++ = '"';
				break;
			default:
				*ThisPtr++ = '\'';
				break;
			}
			break;

		//
		// Double Quote Mark
		//
		case '"':
			switch(qmark)
			{
			case '"':
				qmark = '\0';
				*ThisPtr++ = '"';
				break;
			case '\0':
				qmark = '"';
				*ThisPtr++ = '"';
				break;
			default:
				*ThisPtr++ = '\\';
				*ThisPtr++ = '"';
				break;
			}
			break;

		case 'B':
		case 'b':
		case 'W':
		case 'w':
		case 'L':
		case 'l':
		case 'f':
		case 'F':
		case 'd':
		case 'D':
		case 'g':
		case 'G':
		case 'h':
		case 'H':
		case 'c':
		case 'C':
			if (qmark == '\0')
			{
				*ThisPtr = '\0';
				mangle_string_value(ThisPart, Base, *Text);
				ThisPtr = ThisPart + strlen(ThisPart);
			}
			else
			{
				*ThisPtr++ = *Text;
			}
			break;

		default:
			*ThisPtr++ = *Text;
			break;
		}

		Text++;
	}

	//
	// Return result
	//
	*ThisPtr = '\0';
	return (ThisPart);
}

/*******************************************************************************

Local Function:

	mangle_string_value()

Title:

	Prints out a number in a given base.

Description:

	This function processes a quoted string

Bugs:


*******************************************************************************/

static void mangle_string_value(char *ThisPart, int Base, char Type)
{
	char *bascode;
	int len;
	char worktext[64];
	int StartChar = 1;

	assert(ThisPart != NULL);

	//
	// Set up format for outputing of this number
	//
	switch(Base)
	{
	case 2:
		bascode = "0b%s";
		break;
	case 8:
		bascode = "0%s";
		break;
	case 16:
		bascode = "x%s";
		break;
	default: // 10
		bascode = "%s";
		break;
	}

	//
	// Lose final quote mark
	//
	len = strlen(ThisPart);
	if (len > 0)
	{
		ThisPart[len-1] = '\0';
	}

	/*
	 * Skip over leading zero's
	 */
	while(ThisPart[StartChar] == '0')
	{
		StartChar++;
	}
	if ((isdigit(ThisPart[StartChar]) == 0) && (StartChar != 1))
	{
		StartChar--;
	}

	switch(Type)
	{
	case 'B':
	case 'b':
	case 'W':
	case 'w':
		sprintf(worktext, bascode, ThisPart+StartChar);
		break;

	case 'L':
	case 'l':
		sprintf(worktext, bascode, ThisPart+StartChar);
		strcat(worktext, "L");
		break;

	case 'C':
	case 'c':
		if (Base == 10)
		{
			long value = atoi(ThisPart + 1);
			if (value == 0)
			{
				strcpy(worktext, "BString('\\0')");
				break;
			}
		}
		worktext[0] = '"';
		worktext[1] = '\\';
		if (Base == 10)
		{
			long value = atoi(ThisPart + 1);
			sprintf(worktext + 2, "x%x", value);
		}
		else
		{
			sprintf(worktext + 2, bascode, ThisPart + 1);
		}
		strcat(worktext, "\"");
		break;

	default:
		strcpy(worktext, ThisPart + 1);
		break;

	}

	strcpy(ThisPart, worktext);
}

