/*******************************************************************************

Program:

	Main.cc
	01.04

Title:

	Basic Translater Main Function

Description:

	This is the main driver function for the basic translater.

Algorithm:

	This translater operates in steps on the source code:
	<< This sequence has been changed somewhat >>

	1. Rough translation using LEX and YACC to create a tree out
	   of the entire program. This tree only approximates the actual
	   program sequence.

	2. Scan through tree, removing out initilization code, and making
	   tables of line numbers and variables used in the program.

	3. Write out the variable definition section of the output.

	4. Modify the tree to actual execution format, converting
	   if-then-else, for-next, ... into a better node format.

	5. Lose line numbers and labels that are not used.

	6. Pull subroutines out into seperate units.

	7. Optimize tree. See if we can replace statements with
		auto-increment/decrement, ...

	8. Write out code section of program.

Bugs:

	Requires the source code to be formatted into a more strict
	format than the actual basic compiler likes.

History:

	01.01.00 - 08/30/1996 - Kevin Handy
		Set version number to 1.1, and install on internet.

	01.01.01 - 08/31/1996 - Kevin Handy
		Changed 'TreeStruct' to 'Node', which makes more sense.
		Modified for privatization of Nodes data.
		Lose extra calls to 'UnDownLink', which werent needed.
		Found bug in "IF x THEN IF y THEN z else a else b"

	01.01.02 - 10/22/1996 - Kevin Handy
		Added variables 'IntegerType' and 'RealType'.

	01.02.01 - 12/17/1996 - Kevin Handy
		Move initalization of 'IntegerType' and 'RealType' from
		parse.y to main().

	01.02.02 - 01/05/1997 - Kevin Handy
		Created a function to write out the titles.
		Added date and time of translation.

	01.03.01 - 03/04/1997 - Kevin Handy
		Use getopt to parse command line

	01.03.02 - 04/02/1997 - Kevin Handy
		Handle UNLESS statelents like IF statements.

	01.04.00 - 06/16/1997 - Kevin Handy
		Set version number to 1.4

	01.04.01 - 07/25/1997 - Kevin Handy
		Add 'line' option to leave all line numbers on.

*******************************************************************************/

//
// System Include Files
//
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <time.h>
#include <unistd.h>
#include <getopt.h>

//
// Project Include Files
//
#include "basic.h"
#include "nodes.h"
#include "parse.h"

//
// Module #defined Constants
//

//
// Module #defined Functions
//

//
// Global Variables
//
VariableList *Variables;

int CompileFlag;

int KeepAllLines = FALSE;
int NeedIostreamH = 0;
int NeedStringH = 0;
int NeedMathH = 0;
int NeedErrorH = 0;
int NeedStdlibH = 0;
int NeedRFA = 0;
int NeedBasicFun = 0;
int NeedTimeH = 0;
int NeedDataList = 0;
int NeedPuse = 0;
int NeedChannel = 0;
int NeedUnistd = 0;

int IntegerType = VARTYPE_LONG;
int RealType = VARTYPE_DOUBLE;

int DebugDump = FALSE;
int DebugDumpOne = FALSE;
int VariableDump = FALSE;
int PositionDump = FALSE;


//
// Module Structure Definitions
//

//
// Module Function Prototypes
//
static void WriteHeader(ostream& os, const char* Name);

//
// Module Variables
//
ostream *OutFile;		// Only global because YACC code
				// can't pass this easily to DpProgram

static const char* OptionList = "t:T:vVcCpPo:O:hHl";

#ifdef USE_LONGOPT
static const struct option OptionListLong[] =
{
	{"trace", 1, 0, 't'},
	{"varlist", 0, 0, 'v'},
	{"compile", 0, 0, 'c'},
	{"position", 0, 0, 'p'},
	{"output", 1, 0, 'o'},
	{"help", 0, 0, 'h'},
	{"line", 0, 0, 'l'},
	{0, 0, 0, 0}
};
#endif


/*******************************************************************************

Global Function:

	main()

Title:

	Main driver function.

Description:

	This is the overall controlling function of the translater.

Bugs:

	Needs command line options.

*******************************************************************************/


main(int argc, char *argv[])
{
	//
	// Local Variables
	//
	int DidOneFlag = 0;
	int OptionSelected;
	int OptionIndex;

	//
	// Initialize variables
	//
	OutFile = &cout;
	CompileFlag = 0;
	KeepAllLines = FALSE;

	//
	// Handle input arguments
	//
#ifdef USE_LONGOPT
	while ((OptionSelected = getopt_long(argc, argv, OptionList,
		OptionListLong, &OptionIndex)) != -1)
#else
	while ((OptionSelected = getopt(argc, argv, OptionList)) != -1)
#endif
	{
		//
		// Process one arguement
		//
		switch (OptionSelected)
		{
		case 't':
		case 'T':
			//
			// Tracing
			//
			switch(optarg[0])
			{
			case '1':
				DebugDump = TRUE;
				break;
			case '2':
				DebugDumpOne = TRUE;
				break;
			default:
				DebugDump = TRUE;
				DebugDumpOne = TRUE;
				break;
			}
			break;

		case 'v':
		case 'V':
			//
			// Variable dump(s)
			//
			VariableDump = TRUE;
			break;

		case 'o':
		case 'O':
			//
			// Output File
			//
			if (OutFile != &cout)
			{
				delete OutFile;
				OutFile = NULL;
			}

			OutFile = new ofstream(optarg);
			if (OutFile == NULL)
			{
				cerr << "Unable to open " << optarg <<
					"for output" << endl;
				exit(1);
			}

			break;

		case 'h':
		case 'H':
			DidOneFlag = 1;
			cerr << "Usage: basic [-h] [-t<n>] [-v] [-c] [-p] [-l] "
				"[-o <file>] <source>" << endl;
		cerr << "   or: basic [--help] [--trace <n>] [--varlist] " << endl <<
			"       [--compile] [--position] [--lines] [--output <file>] <source>" << endl;

			break;

		case 'c':
		case 'C':
			CompileFlag = 1;
			break;

		case 'p':
		case 'P':
			PositionDump = TRUE;
			break;

		case 'l':
		case 'L':
			KeepAllLines = TRUE;
			break;

		default:
			cerr << "%Error on command line '-h' for help" << endl;
			exit(EXIT_FAILURE);
		}
	}

	//
	// Anything left over is files names to process
	//
	while (optind < argc)
	{
		//
		// At least one file name on command line
		//
		DidOneFlag = 1;

		//
		// Reset line counter
		//
		xline = 1;
		include_stack_pointer = 0;

		//
		// Open up source file
		//
		if ((yyin = fopen(argv[optind], "r")) == 0)
		{
			cerr << "Unable to open input file " <<
				argv[optind] << endl;
			exit(1);
		}

		WriteHeader(*OutFile, argv[optind]);

		//
		// Step 1. Parse the program. YACC will call the
		//  rest of the program if it scans correctly.
		//
		if (PositionDump)
		{
			cerr << "Parsing." << endl;
		}

		Variables = new VariableList;

		/*
		 * Default variable sizes
		 */
		IntegerType = VARTYPE_LONG;
		RealType = VARTYPE_DOUBLE;

		if (yyparse() != 0)
		{
			cerr << "%Failure during parse" << endl;
			delete Variables;
			fclose(yyin);
			exit(1);
		}

		//
		// Close the input file
		//
		fclose(yyin);
		delete Variables;

		optind++;
	}

	//
	// If we haven't seen any file names, then get input
	// from the users terminal
	//
	if (DidOneFlag == 0)
	{
		//
		// Reset line counter
		//
		xline = 1;
		include_stack_pointer = 0;

		//
		// Read from standard input if we haven't handled
		// any input files yet.
		//
		yyin = stdin;

		WriteHeader(*OutFile, "Standard Input");

		if (PositionDump)
		{
			cerr << "Parsing." << endl;
		}

		Variables = new VariableList;

		if (yyparse() != 0)
		{
			cerr << "%Failure during parse" << endl;
			delete Variables;
			exit(1);
		}

		//
		// Close the input file
		//
		delete Variables;
	}

	if (OutFile != &cout)
	{
		delete OutFile;
	}

	return 0;
}

/*******************************************************************************

Local function:

	WriteHeader()

Title:

	Write Header To Output File

Description:

	Writes out a nice little title at the start if an output
	translation saying I'll take the blame for the lousy code
	output.
*******************************************************************************/

static void WriteHeader(ostream& os, const char* Name)
{
	time_t timer;
	struct tm *tblock;
	char Text[64];

	timer = time(NULL);
	tblock = localtime(&timer);

	strftime(Text, sizeof(Text), "%A, %B %d, %Y at %X",
		tblock);

	os << "//" << endl;
	os << "// Source: " << Name << endl;
	os << "// Translated from Basic to C++ using btran(~tm:)" << endl;
	os << "// on " << Text << endl;
	os << "//" << endl << endl;
}


#ifndef __MSDOS__

/*******************************************************************************

Global Function:

	strupr()

Title:

	Convert a string to upper case

Description:

	This function will convert a text string to upper case.

Bugs:

	This function exists in MSDOS, but not in ANSI-C definition,
	and thus is defined here to make it available.

*******************************************************************************/

char *strupr(char *s)
{
	char *s1 = s;

	while (*s1)
	{
		*s1 = toupper(*s1);
		s1++;
	}

	return s;
}

/*******************************************************************************

Global Function:

	strlwr()

Title:

	Convert a string to lower case

Description:

	This function will convert a text string to lower case.

Bugs:

	This function exists in MSDOS, but not in ANSI-C definition,
	and thus is defined here to make it available.

*******************************************************************************/

char *strlwr(char *s)
{
	char *s1 = s;

	while (*s1)
	{
		*s1 = tolower(*s1);
		s1++;
	}

	return s;
}
#endif
