/*******************************************************************************

Program:

	variable.cc
	01.04

Title:

	Routines to handle variables.

Description:

	This file contains routines to handle the variable definitions
	necessary for basic.

Bugs:

History:

	01.01.00 - 08/30/1996 - Kevin Handy
		Update version number to 1.1, and update on internet.

	01.01.01 - 10/22/1996 - Kevin Handy
		Handle new 'IntegerType' and 'RealType' variables.

	01.03.01 - 04/30/1997 - Kevin Handy
		Guestimate the number of parameters to a array/function,
		and use this guess when outputing default definitions.

	01.04.00 - 06/16/1997 - Kevin Handy
		Set version to 1.6

	01.04.01 - 06/24/1997 - Kevin Handy
		Output correct for 'VARTYPE_REAL' based on 'option' value

	01.04.02 - 07/06/1997 - Kevin Handy
		Broke out some common code into the functions DumpVarType,
		OutputVarType, and GuessVarType.

*******************************************************************************/

/*
 * System Include Files
 */
#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

//
// Project Include Files
//
#include "variable.h"
#include "basic.h"
#include "parse.h"


/*******************************************************************************

Member Function:

	VariableStruct()

Title:

	Constructs a variable item.

Description:

	This routine fill add a variable to the variable table.

Bugs:

*******************************************************************************/

VariableStruct::VariableStruct
(
	char *VarName,
	int newType,
	int newClass,
	int newOutput,
	int newParCount
)
{
	BasicName = NULL;
	Type = newType;
	Class = newClass;
	CName = NULL;
	Output = newOutput;
	ParCount = newParCount;
	Left = NULL;
	Right = NULL;
	EverUsed = 0;

	//
	// Allocate space for variable name
	//
	if (VarName)
	{
		char* ThisName = new char[strlen(VarName)+1];
		strcpy(ThisName, VarName);
		BasicName = ThisName;
	}

}

/*******************************************************************************

Member Function:

	~VariableStruct()

Title:

	Destruct a variable.

Description:

	Destroys variables.

Bugs:

*******************************************************************************/

VariableStruct::~VariableStruct()
{
	delete BasicName;
	delete CName;
}

/*******************************************************************************

Member Function:

	Dump()

Title:

	Print out variable.

Description:

	Prints out a variables information.

Bugs:

	Ugly output.

*******************************************************************************/

void VariableStruct::Dump()
{
	//
	// Print out this variable
	//
	cout << "Var: " << BasicName;

	if (Class == VARCLASS_ARRAY)
	{
		cout << "[" << ParCount << "]";
	}

	cout << ", Type " << Type << DumpVarType(Type) <<
		", Class " << Class;

	switch (Class)
	{
	case VARCLASS_NONE:
		cout << " Regular variable";
		break;

	case VARCLASS_ARRAY:
		cout << " An Array";
		break;

	case VARCLASS_FUNC:
		cout << " Function";
		break;

	case VARCLASS_CONST:
		cout << " Constant";
		break;

	default:
		cout << " Undefined";
		break;
	}

	if (CName != 0)
	{
		cout << ", CName " << CName;

		if (Class == VARCLASS_ARRAY)
		{
			cout << "[]";
		}
	}

	cout << endl;
}


/*******************************************************************************

Member Function:

	VariableStruct::OutputDef()

Title:

	Dump out one variable node.

Description:

	This function will output one variables definition to the
	output file. These are only used if the definition isn't output
	as part of the source code.

Bugs:


*******************************************************************************/

void VariableStruct::OutputDef
(
	ostream& os,
	int Level
)
{
	//
	// Local Variables
	//
	char* OldSep;
	int ActualType = Type;

	//
	// Skip if already output
	//
	if (Output == TRUE)
	{
		return;
	}

	//
	// Tab over
	//
	Indent(os, Level);

	//
	// Validation Check
	//
	assert(BasicName != 0);

	//
	// Output This Variable
	//
	switch (Class)
	{
	case VARCLASS_CONST:
		os << "const ";
		break;
	}

	//
	// Change "Integer" and "Real" to more exact type
	//
	switch(ActualType)
	{
	case VARTYPE_INTEGER:
		ActualType = IntegerType;
		break;

	case VARTYPE_REAL:
		ActualType = RealType;
		break;
	}

	os << OutputVarType(ActualType) << " ";

	//
	// Pick a good name
	//
	if (CName != 0)
	{
		os << CName;
	}
	else
	{
		os << BasicName;
	}

	switch(Type)
	{
	case VARTYPE_FIXSTR:
		os << "[]";
		break;
	}


	//
	// Add brackets if this is an array definition
	//
	switch (Class)
	{
	case VARCLASS_ARRAY:
		//
		// Default array dimension under Basic is 0..10
		// for most Basic's I've seen.
		//
		os << "[11]";
		for (int loop = 2; loop <= ParCount; loop++)
		{
			os << "[11]";
		}
		break;

	case VARCLASS_FUNC:
		os << "()";
		break;
	}

	os << ";" << endl;

	//
	// Flag this one as having been output.
	//
	Output = TRUE;
}

/*******************************************************************************

Local Function:

	Indent()

Title:

	Indents to current level.

Description:

	This function is used to indent the code.

Bugs:


*******************************************************************************/

void VariableStruct::Indent(ostream& os, int Level)
{
	int TabLoop;

	for (TabLoop = 1; TabLoop <= Level; TabLoop++)
	{
		os << "\t";
	}
}

/*******************************************************************************

Member Function:

	SetCName()

Title:

	Sets up the C Variable Name.

Description:

	Assigned the C variable name.

Bugs:


*******************************************************************************/

void VariableStruct::SetCName(const char* NewName)
{
	//
	// Lose old value
	//
	delete CName;

	//
	// Move over new value
	//
	CName = new char[strlen(NewName) + 1];
	strcpy(CName, NewName);
}
/*******************************************************************************

Member Function:

        GuessVarType()

Title:

        Guess the type of a variable based on it's name

Description:

        Looks at a variables name and attempts to guess what type
	of variable it is. This type of guess is only used when
	an actual definition is not given.

Bugs:

        Really simple guesses.

*******************************************************************************/

int GuessVarType(const char* Name)
{
	switch(Name[strlen(Name) - 1])
	{
	case '$':
		return VARTYPE_DYNSTR;
	case '%':
		return VARTYPE_INTEGER;
	default:
		return VARTYPE_REAL;
	}
}

/*******************************************************************************

Member Function:

	DumpType()

Title:

	Print out variable type.

Description:

	Prints out a variable type.

Bugs:

	Ugly output.

*******************************************************************************/

const char* DumpVarType(int VarType)
{
	char* result;

	switch (VarType)
	{
	case VARTYPE_DYNSTR:
		result = " Dynamic String";
		break;

	case VARTYPE_FIXSTR:
		result = " Fixed length string";
		break;

	case VARTYPE_WORD:
		result = " 16 bit integer";
		break;

	case VARTYPE_INTEGER:
		result = " integer";
		break;

	case VARTYPE_LONG:
		result = " 32 bit integer";
		break;

	case VARTYPE_REAL:
		result = " float";
		break;

	case VARTYPE_DOUBLE:
		result = " double";
		break;

	case VARTYPE_STRUCT:
		result = " Structure";
		break;

	case VARTYPE_VOID:
		result = " Void";
		break;

	case VARTYPE_BYTE:
		result = " Byte";
		break;

	case VARTYPE_RFA:
		result = " RFA";
		break;

	case VARTYPE_LABEL:
		result = " Label";
		break;

	case VARTYPE_DECIMAL:
		result = " Decimal";
		break;

	case VARTYPE_SINGLE:
		result = " Single";
		break;

	case VARTYPE_GFLOAT:
		result = " GFloat";
		break;

	case VARTYPE_HFLOAT:
		result = " HFloat";
		break;

	default:
		result = " Undefined Type";
		break;
	}
	return result;
}

/*******************************************************************************

Member Function:

	OutputVarType()

Title:

	Print out variable type.

Description:

	Prints out a variable type.

Bugs:

	Ugly output.

*******************************************************************************/

const char* OutputVarType(int VarType)
{
	int ActualType = VarType;
	char* result;

	//
	// Change "Integer" and "Real" to more exact type
	//
	switch(ActualType)
	{
	case VARTYPE_INTEGER:
		ActualType = IntegerType;
		break;

	case VARTYPE_REAL:
		ActualType = RealType;
		break;
	}

	switch(ActualType)
	{
	case VARTYPE_DYNSTR:
		result = "BString";
		break;

	case VARTYPE_FIXSTR:
		result = "char";
		break;

	case VARTYPE_INTEGER:
		result = "int";
		break;

	case VARTYPE_WORD:
		result = "Int";
		break;

	case VARTYPE_LONG:
		result = "Int32";
		break;

	case VARTYPE_REAL:
		switch(RealType)
		{
		case VARTYPE_DOUBLE:
			result = "Dfloat";
			break;

		case VARTYPE_GFLOAT:
			result = "Gfloat";
			break;

		case VARTYPE_HFLOAT:
			result = "Hfloat";
			break;
		}
		break;

	case VARTYPE_DOUBLE:
		result = "Dfloat";
		break;

	case VARTYPE_VOID:
		result = "void";
		break;

	case VARTYPE_BYTE:
		result = "char";
		break;

	case VARTYPE_RFA:
		result = "RFA";
		break;

	case VARTYPE_DECIMAL:
		result = "Decimal";
		break;

	case VARTYPE_SINGLE:
		result = "Sfloat";
		break;

	case VARTYPE_GFLOAT:
		result = "Gfloat";
		break;

	case VARTYPE_HFLOAT:
		result = "Hfloat";
		break;

	default:
		result = "??unknown??";
		break;

	}
	return result;
}

